# Auto-generated. Do not edit!
'''
The ``measurements`` module provides functions that apply calibrated noise to data to ensure differential privacy.
For more context, see :ref:`measurements in the User Guide <measurements-user-guide>`.

For convenience, all the functions of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The methods of this module will then be accessible at ``dp.m``.
'''
from deprecated.sphinx import deprecated # noqa: F401 (Not every file actually has deprecated functions.)

from opendp._convert import *
from opendp._lib import *
from opendp.mod import *
from opendp.typing import *
from opendp.typing import _substitute # noqa: F401
from opendp.core import *
from opendp.domains import *
from opendp.metrics import *
from opendp.measures import *
__all__ = [
    "debias_randomized_response_bitvec",
    "make_alp_queryable",
    "make_canonical_noise",
    "make_gaussian",
    "make_gaussian_threshold",
    "make_geometric",
    "make_laplace",
    "make_laplace_threshold",
    "make_noise",
    "make_noise_threshold",
    "make_noisy_max",
    "make_noisy_top_k",
    "make_private_expr",
    "make_private_lazyframe",
    "make_private_quantile",
    "make_randomized_response",
    "make_randomized_response_bitvec",
    "make_randomized_response_bool",
    "make_report_noisy_max_gumbel",
    "make_user_measurement",
    "then_alp_queryable",
    "then_canonical_noise",
    "then_gaussian",
    "then_gaussian_threshold",
    "then_geometric",
    "then_laplace",
    "then_laplace_threshold",
    "then_noise",
    "then_noise_threshold",
    "then_noisy_max",
    "then_noisy_top_k",
    "then_private_expr",
    "then_private_lazyframe",
    "then_private_quantile",
    "then_randomized_response_bitvec",
    "then_report_noisy_max_gumbel",
    "then_user_measurement"
]


def debias_randomized_response_bitvec(
    answers,
    f: float
):
    r"""Convert a vector of randomized response bitvec responses to a frequency estimate

    Computes the sum of the answers into a $k$-length vector $Y$ and returns
    $Y\frac{Y-\frac{f}{2}}{1-f}$


    Required features: `contrib`

    [debias_randomized_response_bitvec in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.debias_randomized_response_bitvec.html)

    .. end-markdown

    :param answers: A vector of BitVectors with consistent size
    :param f: The per bit flipping probability used to encode ``answers``
    :type f: float
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_answers = py_to_c(answers, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=["BitVector"]))
    c_f = py_to_c(f, c_type=ctypes.c_double, type_name="f64")

    # Call library function.
    lib_function = lib.opendp_measurements__debias_randomized_response_bitvec
    lib_function.argtypes = [AnyObjectPtr, ctypes.c_double]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_answers, c_f), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': 'debias_randomized_response_bitvec',
            '__module__': 'measurements',
            '__kwargs__': {
                'answers': answers, 'f': f
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_alp_queryable(
    input_domain: Domain,
    input_metric: Metric,
    scale: float,
    total_limit,
    value_limit = None,
    size_factor = 50,
    alpha = 4
) -> Measurement:
    r"""Measurement to release a queryable containing a DP projection of bounded sparse data.

    The size of the projection is O(total * size_factor * scale / alpha).
    The evaluation time of post-processing is O(beta * scale / alpha).

    `size_factor` is an optional multiplier (defaults to 50) for setting the size of the projection.
    There is a memory/utility trade-off.
    The value should be sufficiently large to limit hash collisions.


    Required features: `contrib`

    [make_alp_queryable in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_alp_queryable.html)

    **Citations:**

    * [ALP21 Differentially Private Sparse Vectors with Low Error, Optimal Space, and Fast Access](https://arxiv.org/abs/2106.10068) Algorithm 4

    **Supporting Elements:**

    * Input Domain:   `MapDomain<AtomDomain<K>, AtomDomain<CI>>`
    * Output Type:    `L0PInfDistance<1, AbsoluteDistance<CI>>`
    * Input Metric:   `MaxDivergence`
    * Output Measure: `Queryable<K, f64>`

    .. end-markdown

    :param input_domain: Domain of input data
    :type input_domain: Domain
    :param input_metric: Metric on input domain
    :type input_metric: Metric
    :param scale: Privacy loss parameter. This is equal to epsilon/sensitivity.
    :type scale: float
    :param total_limit: Either the true value or an upper bound estimate of the sum of all values in the input.
    :param value_limit: Upper bound on individual values (referred to as β). Entries above β are clamped.
    :param size_factor: Optional multiplier (default of 50) for setting the size of the projection.
    :param alpha: Optional parameter (default of 4) for scaling and determining p in randomized response step.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # Standardize type arguments.
    CI = get_value_type(get_carrier_type(input_domain)) # type: ignore

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_total_limit = py_to_c(total_limit, c_type=ctypes.c_void_p, type_name=CI)
    c_value_limit = py_to_c(value_limit, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=[CI]))
    c_size_factor = py_to_c(size_factor, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=["u32"]))
    c_alpha = py_to_c(alpha, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=["u32"]))

    # Call library function.
    lib_function = lib.opendp_measurements__make_alp_queryable
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_scale, c_total_limit, c_value_limit, c_size_factor, c_alpha), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_alp_queryable',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'scale': scale, 'total_limit': total_limit, 'value_limit': value_limit, 'size_factor': size_factor, 'alpha': alpha
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_alp_queryable(
    scale: float,
    total_limit,
    value_limit = None,
    size_factor = 50,
    alpha = 4
):  
    r"""partial constructor of make_alp_queryable

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_alp_queryable`

    :param scale: Privacy loss parameter. This is equal to epsilon/sensitivity.
    :type scale: float
    :param total_limit: Either the true value or an upper bound estimate of the sum of all values in the input.
    :param value_limit: Upper bound on individual values (referred to as β). Entries above β are clamped.
    :param size_factor: Optional multiplier (default of 50) for setting the size of the projection.
    :param alpha: Optional parameter (default of 4) for scaling and determining p in randomized response step.
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_alp_queryable(
        input_domain=input_domain,
        input_metric=input_metric,
        scale=scale,
        total_limit=total_limit,
        value_limit=value_limit,
        size_factor=size_factor,
        alpha=alpha))
    output.__opendp_dict__ = {
            '__function__': 'then_alp_queryable',
            '__module__': 'measurements',
            '__kwargs__': {
                'scale': scale, 'total_limit': total_limit, 'value_limit': value_limit, 'size_factor': size_factor, 'alpha': alpha
            },
        }
    return output



def make_canonical_noise(
    input_domain: Domain,
    input_metric: Metric,
    d_in: float,
    d_out: tuple[Any, Any]
) -> Measurement:
    r"""Make a Measurement that adds noise from a canonical noise distribution.
    The implementation is tailored towards approximate-DP,
    resulting in noise sampled from the Tulap distribution.


    Required features: `contrib`

    [make_canonical_noise in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_canonical_noise.html)

    **Citations:**

    - [AV23 Canonical Noise Distributions and Private Hypothesis Tests](https://projecteuclid.org/journals/annals-of-statistics/volume-51/issue-2/Canonical-noise-distributions-and-private-hypothesis-tests/10.1214/23-AOS2259.short)

    **Supporting Elements:**

    * Input Domain:   `AtomDomain<f64>`
    * Output Type:    `AbsoluteDistance<f64>`
    * Input Metric:   `Approximate<MaxDivergence>`
    * Output Measure: `f64`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/canonical_noise/make_canonical_noise.pdf)

    .. end-markdown

    :param input_domain: Domain of the input.
    :type input_domain: Domain
    :param input_metric: Metric of the input.
    :type input_metric: Metric
    :param d_in: Sensitivity
    :type d_in: float
    :param d_out: Privacy parameters (ε, δ)
    :type d_out: tuple[Any, Any]
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_d_in = py_to_c(d_in, c_type=ctypes.c_double, type_name="f64")
    c_d_out = py_to_c(d_out, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Tuple', args=["f64", "f64"]))

    # Call library function.
    lib_function = lib.opendp_measurements__make_canonical_noise
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_d_in, c_d_out), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_canonical_noise',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'd_in': d_in, 'd_out': d_out
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_canonical_noise(
    d_in: float,
    d_out: tuple[Any, Any]
):  
    r"""partial constructor of make_canonical_noise

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_canonical_noise`

    :param d_in: Sensitivity
    :type d_in: float
    :param d_out: Privacy parameters (ε, δ)
    :type d_out: tuple[Any, Any]
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_canonical_noise(
        input_domain=input_domain,
        input_metric=input_metric,
        d_in=d_in,
        d_out=d_out))
    output.__opendp_dict__ = {
            '__function__': 'then_canonical_noise',
            '__module__': 'measurements',
            '__kwargs__': {
                'd_in': d_in, 'd_out': d_out
            },
        }
    return output



def make_gaussian(
    input_domain: Domain,
    input_metric: Metric,
    scale: float,
    k = None,
    MO: RuntimeTypeDescriptor = "ZeroConcentratedDivergence"
) -> Measurement:
    r"""Make a Measurement that adds noise from the Gaussian(`scale`) distribution to the input.

    Valid inputs for `input_domain` and `input_metric` are:

    | `input_domain`                  | input type   | `input_metric`          |
    | ------------------------------- | ------------ | ----------------------- |
    | `atom_domain(T)`                | `T`          | `absolute_distance(QI)` |
    | `vector_domain(atom_domain(T))` | `Vec<T>`     | `l2_distance(QI)`       |


    Required features: `contrib`

    [make_gaussian in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_gaussian.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `DI::Carrier`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/noise/distribution/gaussian/make_gaussian.pdf)

    .. end-markdown

    :param input_domain: Domain of the data type to be privatized.
    :type input_domain: Domain
    :param input_metric: Metric of the data type to be privatized.
    :type input_metric: Metric
    :param scale: Noise scale parameter for the gaussian distribution. ``scale`` == standard_deviation.
    :type scale: float
    :param k: The noise granularity in terms of 2^k.
    :param MO: Output Measure. The only valid measure is ``ZeroConcentratedDivergence``.
    :type MO: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.enable_features('contrib')
    >>> input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    >>> gaussian = dp.m.make_gaussian(*input_space, scale=1.0)
    >>> print('100?', gaussian(100.0))
    100? ...

    Or, more readably, define the space and then chain:

    >>> gaussian = input_space >> dp.m.then_gaussian(scale=1.0)
    >>> print('100?', gaussian(100.0))
    100? ...

    """
    assert_features("contrib")

    # Standardize type arguments.
    MO = RuntimeType.parse(type_name=MO)

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_k = py_to_c(k, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=["i32"]))
    c_MO = py_to_c(MO, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_measurements__make_gaussian
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, ctypes.c_void_p, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_scale, c_k, c_MO), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_gaussian',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'scale': scale, 'k': k, 'MO': MO
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_gaussian(
    scale: float,
    k = None,
    MO: RuntimeTypeDescriptor = "ZeroConcentratedDivergence"
):  
    r"""partial constructor of make_gaussian

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_gaussian`

    :param scale: Noise scale parameter for the gaussian distribution. ``scale`` == standard_deviation.
    :type scale: float
    :param k: The noise granularity in terms of 2^k.
    :param MO: Output Measure. The only valid measure is ``ZeroConcentratedDivergence``.
    :type MO: :py:ref:`RuntimeTypeDescriptor`

    :example:

    >>> dp.enable_features('contrib')
    >>> input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    >>> gaussian = dp.m.make_gaussian(*input_space, scale=1.0)
    >>> print('100?', gaussian(100.0))
    100? ...

    Or, more readably, define the space and then chain:

    >>> gaussian = input_space >> dp.m.then_gaussian(scale=1.0)
    >>> print('100?', gaussian(100.0))
    100? ...

    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_gaussian(
        input_domain=input_domain,
        input_metric=input_metric,
        scale=scale,
        k=k,
        MO=MO))
    output.__opendp_dict__ = {
            '__function__': 'then_gaussian',
            '__module__': 'measurements',
            '__kwargs__': {
                'scale': scale, 'k': k, 'MO': MO
            },
        }
    return output



def make_gaussian_threshold(
    input_domain: Domain,
    input_metric: Metric,
    scale: float,
    threshold,
    k = None,
    MO: RuntimeTypeDescriptor = "Approximate<ZeroConcentratedDivergence>"
) -> Measurement:
    r"""Make a Measurement that uses propose-test-release to privatize a hashmap of counts.

    This function takes a noise granularity in terms of 2^k.
    Larger granularities are more computationally efficient, but have a looser privacy map.
    If k is not set, k defaults to the smallest granularity.


    Required features: `contrib`

    [make_gaussian_threshold in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_gaussian_threshold.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `DI::Carrier`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/noise_threshold/distribution/gaussian/make_gaussian_threshold.pdf)

    .. end-markdown

    :param input_domain: Domain of the input.
    :type input_domain: Domain
    :param input_metric: Metric for the input domain.
    :type input_metric: Metric
    :param scale: Noise scale parameter for the laplace distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param threshold: Exclude pairs with values whose distance from zero exceeds this value.
    :param k: The noise granularity in terms of 2^k.
    :param MO: Output Measure.
    :type MO: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # Standardize type arguments.
    MO = RuntimeType.parse(type_name=MO, generics=["TV"])
    TV = get_value_type(get_carrier_type(input_domain)) # type: ignore
    MO = _substitute(MO, TV=TV) # type: ignore

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_threshold = py_to_c(threshold, c_type=ctypes.c_void_p, type_name=TV)
    c_k = py_to_c(k, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=["i32"]))
    c_MO = py_to_c(MO, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_measurements__make_gaussian_threshold
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_scale, c_threshold, c_k, c_MO), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_gaussian_threshold',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'scale': scale, 'threshold': threshold, 'k': k, 'MO': MO
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_gaussian_threshold(
    scale: float,
    threshold,
    k = None,
    MO: RuntimeTypeDescriptor = "Approximate<ZeroConcentratedDivergence>"
):  
    r"""partial constructor of make_gaussian_threshold

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_gaussian_threshold`

    :param scale: Noise scale parameter for the laplace distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param threshold: Exclude pairs with values whose distance from zero exceeds this value.
    :param k: The noise granularity in terms of 2^k.
    :param MO: Output Measure.
    :type MO: :py:ref:`RuntimeTypeDescriptor`
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_gaussian_threshold(
        input_domain=input_domain,
        input_metric=input_metric,
        scale=scale,
        threshold=threshold,
        k=k,
        MO=MO))
    output.__opendp_dict__ = {
            '__function__': 'then_gaussian_threshold',
            '__module__': 'measurements',
            '__kwargs__': {
                'scale': scale, 'threshold': threshold, 'k': k, 'MO': MO
            },
        }
    return output



def make_geometric(
    input_domain: Domain,
    input_metric: Metric,
    scale: float,
    bounds = None,
    MO: RuntimeTypeDescriptor = "MaxDivergence"
) -> Measurement:
    r"""Equivalent to `make_laplace` but restricted to an integer support.
    Can specify `bounds` to run the algorithm in near constant-time.


    Required features: `contrib`

    [make_geometric in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_geometric.html)

    **Citations:**

    * [GRS12 Universally Utility-Maximizing Privacy Mechanisms](https://theory.stanford.edu/~tim/papers/priv.pdf)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `DI::Carrier`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/noise/distribution/geometric/make_geometric.pdf)

    .. end-markdown

    :param input_domain: Domain of the data type to be privatized.
    :type input_domain: Domain
    :param input_metric: Metric of the data type to be privatized.
    :type input_metric: Metric
    :param scale: Noise scale parameter for the distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param bounds: Set bounds on the count to make the algorithm run in constant-time.
    :param MO: Measure used to quantify privacy loss. Valid values are just ``MaxDivergence``
    :type MO: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.enable_features("contrib")
    >>> input_space = dp.atom_domain(T=int), dp.absolute_distance(T=int)
    >>> geometric = dp.m.make_geometric(*input_space, scale=1.0)
    >>> print('100?', geometric(100))
    100? ...

    Or, more readably, define the space and then chain:

    >>> geometric = input_space >> dp.m.then_geometric(scale=1.0)
    >>> print('100?', geometric(100))
    100? ...

    """
    assert_features("contrib")

    # Standardize type arguments.
    MO = RuntimeType.parse(type_name=MO, generics=["T", "OptionT"])
    T = get_atom(get_carrier_type(input_domain)) # type: ignore
    OptionT = RuntimeType(origin='Option', args=[RuntimeType(origin='Tuple', args=[T, T])]) # type: ignore
    MO = _substitute(MO, T=T, OptionT=OptionT) # type: ignore

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_bounds = py_to_c(bounds, c_type=AnyObjectPtr, type_name=OptionT)
    c_MO = py_to_c(MO, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_measurements__make_geometric
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, AnyObjectPtr, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_scale, c_bounds, c_MO), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_geometric',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'scale': scale, 'bounds': bounds, 'MO': MO
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_geometric(
    scale: float,
    bounds = None,
    MO: RuntimeTypeDescriptor = "MaxDivergence"
):  
    r"""partial constructor of make_geometric

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_geometric`

    :param scale: Noise scale parameter for the distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param bounds: Set bounds on the count to make the algorithm run in constant-time.
    :param MO: Measure used to quantify privacy loss. Valid values are just ``MaxDivergence``
    :type MO: :py:ref:`RuntimeTypeDescriptor`

    :example:

    >>> dp.enable_features("contrib")
    >>> input_space = dp.atom_domain(T=int), dp.absolute_distance(T=int)
    >>> geometric = dp.m.make_geometric(*input_space, scale=1.0)
    >>> print('100?', geometric(100))
    100? ...

    Or, more readably, define the space and then chain:

    >>> geometric = input_space >> dp.m.then_geometric(scale=1.0)
    >>> print('100?', geometric(100))
    100? ...

    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_geometric(
        input_domain=input_domain,
        input_metric=input_metric,
        scale=scale,
        bounds=bounds,
        MO=MO))
    output.__opendp_dict__ = {
            '__function__': 'then_geometric',
            '__module__': 'measurements',
            '__kwargs__': {
                'scale': scale, 'bounds': bounds, 'MO': MO
            },
        }
    return output



def make_laplace(
    input_domain: Domain,
    input_metric: Metric,
    scale: float,
    k = None,
    MO: RuntimeTypeDescriptor = "MaxDivergence"
) -> Measurement:
    r"""Make a Measurement that adds noise from the Laplace(`scale`) distribution to the input.

    Valid inputs for `input_domain` and `input_metric` are:

    | `input_domain`                  | input type   | `input_metric`         |
    | ------------------------------- | ------------ | ---------------------- |
    | `atom_domain(T)` (default)      | `T`          | `absolute_distance(T)` |
    | `vector_domain(atom_domain(T))` | `Vec<T>`     | `l1_distance(T)`       |

    Internally, all sampling is done using the discrete Laplace distribution.


    Required features: `contrib`

    [make_laplace in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_laplace.html)

    **Citations:**

    * [GRS12 Universally Utility-Maximizing Privacy Mechanisms](https://theory.stanford.edu/~tim/papers/priv.pdf)
    * [CKS20 The Discrete Gaussian for Differential Privacy](https://arxiv.org/pdf/2004.00010.pdf#subsection.5.2)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `DI::Carrier`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/noise/distribution/laplace/make_laplace.pdf)

    .. end-markdown

    :param input_domain: Domain of the data type to be privatized.
    :type input_domain: Domain
    :param input_metric: Metric of the data type to be privatized.
    :type input_metric: Metric
    :param scale: Noise scale parameter for the Laplace distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param k: The noise granularity in terms of 2^k, only valid for domains over floats.
    :param MO: Measure used to quantify privacy loss. Valid values are just ``MaxDivergence``
    :type MO: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> import opendp.prelude as dp
    >>> dp.enable_features("contrib")
    >>> input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    >>> laplace = dp.m.make_laplace(*input_space, scale=1.0)
    >>> print('100?', laplace(100.0))
    100? ...

    Or, more readably, define the space and then chain:

    >>> laplace = input_space >> dp.m.then_laplace(scale=1.0)
    >>> print('100?', laplace(100.0))
    100? ...

    """
    assert_features("contrib")

    # Standardize type arguments.
    MO = RuntimeType.parse(type_name=MO)

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_k = py_to_c(k, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=["i32"]))
    c_MO = py_to_c(MO, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_measurements__make_laplace
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, ctypes.c_void_p, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_scale, c_k, c_MO), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_laplace',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'scale': scale, 'k': k, 'MO': MO
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_laplace(
    scale: float,
    k = None,
    MO: RuntimeTypeDescriptor = "MaxDivergence"
):  
    r"""partial constructor of make_laplace

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_laplace`

    :param scale: Noise scale parameter for the Laplace distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param k: The noise granularity in terms of 2^k, only valid for domains over floats.
    :param MO: Measure used to quantify privacy loss. Valid values are just ``MaxDivergence``
    :type MO: :py:ref:`RuntimeTypeDescriptor`

    :example:

    >>> import opendp.prelude as dp
    >>> dp.enable_features("contrib")
    >>> input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    >>> laplace = dp.m.make_laplace(*input_space, scale=1.0)
    >>> print('100?', laplace(100.0))
    100? ...

    Or, more readably, define the space and then chain:

    >>> laplace = input_space >> dp.m.then_laplace(scale=1.0)
    >>> print('100?', laplace(100.0))
    100? ...

    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_laplace(
        input_domain=input_domain,
        input_metric=input_metric,
        scale=scale,
        k=k,
        MO=MO))
    output.__opendp_dict__ = {
            '__function__': 'then_laplace',
            '__module__': 'measurements',
            '__kwargs__': {
                'scale': scale, 'k': k, 'MO': MO
            },
        }
    return output



def make_laplace_threshold(
    input_domain: Domain,
    input_metric: Metric,
    scale: float,
    threshold,
    k = None,
    MO: RuntimeTypeDescriptor = "Approximate<MaxDivergence>"
) -> Measurement:
    r"""Make a Measurement that uses propose-test-release to privatize a hashmap of counts.

    This function takes a noise granularity in terms of 2^k.
    Larger granularities are more computationally efficient, but have a looser privacy map.
    If k is not set, k defaults to the smallest granularity.


    Required features: `contrib`

    [make_laplace_threshold in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_laplace_threshold.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `DI::Carrier`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/noise_threshold/distribution/laplace/make_laplace_threshold.pdf)

    .. end-markdown

    :param input_domain: Domain of the input.
    :type input_domain: Domain
    :param input_metric: Metric for the input domain.
    :type input_metric: Metric
    :param scale: Noise scale parameter for the laplace distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param threshold: Exclude counts that are less than this minimum value.
    :param k: The noise granularity in terms of 2^k.
    :param MO: Output Measure.
    :type MO: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # Standardize type arguments.
    MO = RuntimeType.parse(type_name=MO, generics=["TV"])
    TV = get_value_type(get_carrier_type(input_domain)) # type: ignore
    MO = _substitute(MO, TV=TV) # type: ignore

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_threshold = py_to_c(threshold, c_type=ctypes.c_void_p, type_name=TV)
    c_k = py_to_c(k, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=["i32"]))
    c_MO = py_to_c(MO, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_measurements__make_laplace_threshold
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_scale, c_threshold, c_k, c_MO), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_laplace_threshold',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'scale': scale, 'threshold': threshold, 'k': k, 'MO': MO
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_laplace_threshold(
    scale: float,
    threshold,
    k = None,
    MO: RuntimeTypeDescriptor = "Approximate<MaxDivergence>"
):  
    r"""partial constructor of make_laplace_threshold

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_laplace_threshold`

    :param scale: Noise scale parameter for the laplace distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param threshold: Exclude counts that are less than this minimum value.
    :param k: The noise granularity in terms of 2^k.
    :param MO: Output Measure.
    :type MO: :py:ref:`RuntimeTypeDescriptor`
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_laplace_threshold(
        input_domain=input_domain,
        input_metric=input_metric,
        scale=scale,
        threshold=threshold,
        k=k,
        MO=MO))
    output.__opendp_dict__ = {
            '__function__': 'then_laplace_threshold',
            '__module__': 'measurements',
            '__kwargs__': {
                'scale': scale, 'threshold': threshold, 'k': k, 'MO': MO
            },
        }
    return output



def make_noise(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    scale: float,
    k = None
) -> Measurement:
    r"""Make a Measurement that adds noise from the appropriate distribution to the input.

    Valid inputs for `input_domain` and `input_metric` are:

    | `input_domain`                  | input type   | `input_metric`          |
    | ------------------------------- | ------------ | ----------------------- |
    | `atom_domain(T)`                | `T`          | `absolute_distance(QI)` |
    | `vector_domain(atom_domain(T))` | `Vec<T>`     | `l2_distance(QI)`       |


    Required features: `contrib`

    [make_noise in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_noise.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `DI::Carrier`

    .. end-markdown

    :param input_domain: Domain of the data type to be privatized.
    :type input_domain: Domain
    :param input_metric: Metric of the data type to be privatized.
    :type input_metric: Metric
    :param output_measure: Privacy measure. Either ``MaxDivergence`` or ``ZeroConcentratedDivergence``.
    :type output_measure: Measure
    :param scale: Noise scale parameter.
    :type scale: float
    :param k: The noise granularity in terms of 2^k.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_k = py_to_c(k, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=["i32"]))

    # Call library function.
    lib_function = lib.opendp_measurements__make_noise
    lib_function.argtypes = [Domain, Metric, Measure, ctypes.c_double, ctypes.c_void_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_scale, c_k), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_noise',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'scale': scale, 'k': k
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_noise(
    output_measure: Measure,
    scale: float,
    k = None
):  
    r"""partial constructor of make_noise

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_noise`

    :param output_measure: Privacy measure. Either ``MaxDivergence`` or ``ZeroConcentratedDivergence``.
    :type output_measure: Measure
    :param scale: Noise scale parameter.
    :type scale: float
    :param k: The noise granularity in terms of 2^k.
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_noise(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        scale=scale,
        k=k))
    output.__opendp_dict__ = {
            '__function__': 'then_noise',
            '__module__': 'measurements',
            '__kwargs__': {
                'output_measure': output_measure, 'scale': scale, 'k': k
            },
        }
    return output



def make_noise_threshold(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    scale: float,
    threshold,
    k = None
) -> Measurement:
    r"""Make a Measurement that uses propose-test-release to privatize a hashmap of counts.

    This function takes a noise granularity in terms of 2^k.
    Larger granularities are more computationally efficient, but have a looser privacy map.
    If k is not set, k defaults to the smallest granularity.


    Required features: `contrib`

    [make_noise_threshold in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_noise_threshold.html)

    **Supporting Elements:**

    * Input Domain:   `DI`
    * Output Type:    `MI`
    * Input Metric:   `Approximate<MO>`
    * Output Measure: `DI::Carrier`

    .. end-markdown

    :param input_domain: Domain of the input.
    :type input_domain: Domain
    :param input_metric: Metric for the input domain.
    :type input_metric: Metric
    :param output_measure: Privacy measure. Either ``MaxDivergence`` or ``ZeroConcentratedDivergence``.
    :type output_measure: Measure
    :param scale: Noise scale parameter for the laplace distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param threshold: Exclude counts that are less than this minimum value.
    :param k: The noise granularity in terms of 2^k.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # Standardize type arguments.
    TV = get_value_type(get_carrier_type(input_domain)) # type: ignore

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_threshold = py_to_c(threshold, c_type=ctypes.c_void_p, type_name=TV)
    c_k = py_to_c(k, c_type=ctypes.c_void_p, type_name=RuntimeType(origin='Option', args=["i32"]))

    # Call library function.
    lib_function = lib.opendp_measurements__make_noise_threshold
    lib_function.argtypes = [Domain, Metric, Measure, ctypes.c_double, ctypes.c_void_p, ctypes.c_void_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_scale, c_threshold, c_k), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_noise_threshold',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'scale': scale, 'threshold': threshold, 'k': k
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_noise_threshold(
    output_measure: Measure,
    scale: float,
    threshold,
    k = None
):  
    r"""partial constructor of make_noise_threshold

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_noise_threshold`

    :param output_measure: Privacy measure. Either ``MaxDivergence`` or ``ZeroConcentratedDivergence``.
    :type output_measure: Measure
    :param scale: Noise scale parameter for the laplace distribution. ``scale`` == standard_deviation / sqrt(2).
    :type scale: float
    :param threshold: Exclude counts that are less than this minimum value.
    :param k: The noise granularity in terms of 2^k.
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_noise_threshold(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        scale=scale,
        threshold=threshold,
        k=k))
    output.__opendp_dict__ = {
            '__function__': 'then_noise_threshold',
            '__module__': 'measurements',
            '__kwargs__': {
                'output_measure': output_measure, 'scale': scale, 'threshold': threshold, 'k': k
            },
        }
    return output



def make_noisy_max(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    scale: float,
    negate: bool = False
) -> Measurement:
    r"""Make a Measurement that takes a vector of scores and privately selects the index of the highest score.


    Required features: `contrib`

    [make_noisy_max in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_noisy_max.html)

    **Supporting Elements:**

    * Input Domain:   `VectorDomain<AtomDomain<TIA>>`
    * Output Type:    `LInfDistance<TIA>`
    * Input Metric:   `MO`
    * Output Measure: `usize`

    .. end-markdown

    :param input_domain: Domain of the input vector. Must be a non-nullable ``VectorDomain``
    :type input_domain: Domain
    :param input_metric: Metric on the input domain. Must be ``LInfDistance``
    :type input_metric: Metric
    :param output_measure: One of ``MaxDivergence``, ``ZeroConcentratedDivergence``
    :type output_measure: Measure
    :param scale: Scale for the noise distribution
    :type scale: float
    :param negate: Set to true to return min
    :type negate: bool
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.enable_features("contrib")
    >>> input_space = dp.vector_domain(dp.atom_domain(T=int)), dp.linf_distance(T=int)
    >>> select_index = dp.m.make_noisy_max(*input_space, dp.max_divergence(), scale=1.0)
    >>> print('2?', select_index([1, 2, 3, 2, 1]))
    2? ...

    Or, more readably, define the space and then chain:

    >>> select_index = input_space >> dp.m.then_noisy_max(dp.max_divergence(), scale=1.0)
    >>> print('2?', select_index([1, 2, 3, 2, 1]))
    2? ...

    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_negate = py_to_c(negate, c_type=ctypes.c_bool, type_name="bool")

    # Call library function.
    lib_function = lib.opendp_measurements__make_noisy_max
    lib_function.argtypes = [Domain, Metric, Measure, ctypes.c_double, ctypes.c_bool]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_scale, c_negate), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_noisy_max',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'scale': scale, 'negate': negate
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_noisy_max(
    output_measure: Measure,
    scale: float,
    negate: bool = False
):  
    r"""partial constructor of make_noisy_max

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_noisy_max`

    :param output_measure: One of ``MaxDivergence``, ``ZeroConcentratedDivergence``
    :type output_measure: Measure
    :param scale: Scale for the noise distribution
    :type scale: float
    :param negate: Set to true to return min
    :type negate: bool

    :example:

    >>> dp.enable_features("contrib")
    >>> input_space = dp.vector_domain(dp.atom_domain(T=int)), dp.linf_distance(T=int)
    >>> select_index = dp.m.make_noisy_max(*input_space, dp.max_divergence(), scale=1.0)
    >>> print('2?', select_index([1, 2, 3, 2, 1]))
    2? ...

    Or, more readably, define the space and then chain:

    >>> select_index = input_space >> dp.m.then_noisy_max(dp.max_divergence(), scale=1.0)
    >>> print('2?', select_index([1, 2, 3, 2, 1]))
    2? ...

    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_noisy_max(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        scale=scale,
        negate=negate))
    output.__opendp_dict__ = {
            '__function__': 'then_noisy_max',
            '__module__': 'measurements',
            '__kwargs__': {
                'output_measure': output_measure, 'scale': scale, 'negate': negate
            },
        }
    return output



def make_noisy_top_k(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    k: int,
    scale: float,
    negate: bool = False
) -> Measurement:
    r"""Make a Measurement that takes a vector of scores and privately selects the index of the highest score.


    Required features: `contrib`

    [make_noisy_top_k in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_noisy_top_k.html)

    **Supporting Elements:**

    * Input Domain:   `VectorDomain<AtomDomain<TIA>>`
    * Output Type:    `LInfDistance<TIA>`
    * Input Metric:   `MO`
    * Output Measure: `Vec<usize>`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/noisy_top_k/make_noisy_top_k.pdf)

    .. end-markdown

    :param input_domain: Domain of the input vector. Must be a non-nullable VectorDomain.
    :type input_domain: Domain
    :param input_metric: Metric on the input domain. Must be LInfDistance
    :type input_metric: Metric
    :param output_measure: One of ``MaxDivergence`` or ``ZeroConcentratedDivergence``
    :type output_measure: Measure
    :param k: Number of indices to select.
    :type k: int
    :param scale: Scale for the noise distribution.
    :type scale: float
    :param negate: Set to true to return bottom k
    :type negate: bool
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_k = py_to_c(k, c_type=ctypes.c_size_t, type_name="usize")
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_negate = py_to_c(negate, c_type=ctypes.c_bool, type_name="bool")

    # Call library function.
    lib_function = lib.opendp_measurements__make_noisy_top_k
    lib_function.argtypes = [Domain, Metric, Measure, ctypes.c_size_t, ctypes.c_double, ctypes.c_bool]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_k, c_scale, c_negate), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_noisy_top_k',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'k': k, 'scale': scale, 'negate': negate
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_noisy_top_k(
    output_measure: Measure,
    k: int,
    scale: float,
    negate: bool = False
):  
    r"""partial constructor of make_noisy_top_k

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_noisy_top_k`

    :param output_measure: One of ``MaxDivergence`` or ``ZeroConcentratedDivergence``
    :type output_measure: Measure
    :param k: Number of indices to select.
    :type k: int
    :param scale: Scale for the noise distribution.
    :type scale: float
    :param negate: Set to true to return bottom k
    :type negate: bool
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_noisy_top_k(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        k=k,
        scale=scale,
        negate=negate))
    output.__opendp_dict__ = {
            '__function__': 'then_noisy_top_k',
            '__module__': 'measurements',
            '__kwargs__': {
                'output_measure': output_measure, 'k': k, 'scale': scale, 'negate': negate
            },
        }
    return output



def make_private_expr(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    expr,
    global_scale = None
) -> Measurement:
    r"""Create a differentially private measurement from an [`Expr`].


    Required features: `contrib`, `honest-but-curious`

    [make_private_expr in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_private_expr.html)

    **Why honest-but-curious?:**

    The privacy guarantee governs only at most one evaluation of the released expression.

    **Supporting Elements:**

    * Input Domain:   `WildExprDomain`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `ExprPlan`

    .. end-markdown

    :param input_domain: The domain of the input data.
    :type input_domain: Domain
    :param input_metric: How to measure distances between neighboring input data sets.
    :type input_metric: Metric
    :param output_measure: How to measure privacy loss.
    :type output_measure: Measure
    :param expr: The [``Expr``] to be privatized.
    :param global_scale: A tune-able parameter that affects the privacy-utility tradeoff.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib", "honest-but-curious")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_expr = py_to_c(expr, c_type=AnyObjectPtr, type_name="Expr")
    c_global_scale = py_to_c(global_scale, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=["f64"]))

    # Call library function.
    lib_function = lib.opendp_measurements__make_private_expr
    lib_function.argtypes = [Domain, Metric, Measure, AnyObjectPtr, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_expr, c_global_scale), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_private_expr',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'expr': expr, 'global_scale': global_scale
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_private_expr(
    output_measure: Measure,
    expr,
    global_scale = None
):  
    r"""partial constructor of make_private_expr

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_private_expr`

    :param output_measure: How to measure privacy loss.
    :type output_measure: Measure
    :param expr: The [``Expr``] to be privatized.
    :param global_scale: A tune-able parameter that affects the privacy-utility tradeoff.
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_private_expr(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        expr=expr,
        global_scale=global_scale))
    output.__opendp_dict__ = {
            '__function__': 'then_private_expr',
            '__module__': 'measurements',
            '__kwargs__': {
                'output_measure': output_measure, 'expr': expr, 'global_scale': global_scale
            },
        }
    return output



def make_private_lazyframe(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    lazyframe,
    global_scale = None,
    threshold = None
) -> Measurement:
    r"""Create a differentially private measurement from a [`LazyFrame`].

    Any data inside the [`LazyFrame`] is ignored,
    but it is still recommended to start with an empty [`DataFrame`] and build up the computation using the [`LazyFrame`] API.


    Required features: `contrib`

    [make_private_lazyframe in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_private_lazyframe.html)

    **Supporting Elements:**

    * Input Domain:   `LazyFrameDomain`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `OnceFrame`

    .. end-markdown

    :param input_domain: The domain of the input data.
    :type input_domain: Domain
    :param input_metric: How to measure distances between neighboring input data sets.
    :type input_metric: Metric
    :param output_measure: How to measure privacy loss.
    :type output_measure: Measure
    :param lazyframe: A description of the computations to be run, in the form of a [``LazyFrame``].
    :param global_scale: Optional. A tune-able parameter that affects the privacy-utility tradeoff.
    :param threshold: Optional. Minimum number of rows in each released group.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.enable_features("contrib")
    >>> import polars as pl

    We'll imagine an elementary school is taking a pet census.
    The private census data will have two columns: 

    >>> lf_domain = dp.lazyframe_domain([
    ...     dp.series_domain("grade", dp.atom_domain(T=dp.i32)),
    ...     dp.series_domain("pet_count", dp.atom_domain(T=dp.i32))])

    We also need to specify the column we'll be grouping by.

    >>> lf_domain_with_margin = dp.with_margin(
    ...     lf_domain,
    ...     dp.polars.Margin(
    ...         by=[pl.col("grade")],
    ...         invariant="keys",
    ...         max_length=50))

    With that in place, we can plan the Polars computation, using the ``dp`` plugin. 

    >>> plan = (
    ...     pl.LazyFrame(schema={'grade': pl.Int32, 'pet_count': pl.Int32})
    ...     .group_by("grade")
    ...     .agg(pl.col("pet_count").dp.sum((0, 10), scale=1.0)))

    We now have all the pieces to make our measurement function using `make_private_lazyframe`:

    >>> dp_sum_pets_by_grade = dp.m.make_private_lazyframe(
    ...     input_domain=lf_domain_with_margin,
    ...     input_metric=dp.symmetric_distance(),
    ...     output_measure=dp.max_divergence(),
    ...     lazyframe=plan,
    ...     global_scale=1.0)

    It's only at this point that we need to introduce the private data.

    >>> df = pl.from_records(
    ...     [
    ...         [0, 0], # No kindergarteners with pets.
    ...         [0, 0],
    ...         [0, 0],
    ...         [1, 1], # Each first grader has 1 pet.
    ...         [1, 1],
    ...         [1, 1],
    ...         [2, 1], # One second grader has chickens!
    ...         [2, 1],
    ...         [2, 9]
    ...     ],
    ...     schema=['grade', 'pet_count'], orient="row")
    >>> lf = pl.LazyFrame(df)
    >>> results = dp_sum_pets_by_grade(lf).collect()
    >>> print(results.sort("grade")) # doctest: +ELLIPSIS
    shape: (3, 2)
    ┌───────┬───────────┐
    │ grade ┆ pet_count │
    │ ---   ┆ ---       │
    │ i64   ┆ i64       │
    ╞═══════╪═══════════╡
    │ 0     ┆ ...       │
    │ 1     ┆ ...       │
    │ 2     ┆ ...       │
    └───────┴───────────┘

    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_lazyframe = py_to_c(lazyframe, c_type=AnyObjectPtr, type_name="LazyFrame")
    c_global_scale = py_to_c(global_scale, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=["f64"]))
    c_threshold = py_to_c(threshold, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=["u32"]))

    # Call library function.
    lib_function = lib.opendp_measurements__make_private_lazyframe
    lib_function.argtypes = [Domain, Metric, Measure, AnyObjectPtr, AnyObjectPtr, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_lazyframe, c_global_scale, c_threshold), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_private_lazyframe',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'lazyframe': lazyframe, 'global_scale': global_scale, 'threshold': threshold
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_private_lazyframe(
    output_measure: Measure,
    lazyframe,
    global_scale = None,
    threshold = None
):  
    r"""partial constructor of make_private_lazyframe

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_private_lazyframe`

    :param output_measure: How to measure privacy loss.
    :type output_measure: Measure
    :param lazyframe: A description of the computations to be run, in the form of a [``LazyFrame``].
    :param global_scale: Optional. A tune-able parameter that affects the privacy-utility tradeoff.
    :param threshold: Optional. Minimum number of rows in each released group.

    :example:

    >>> dp.enable_features("contrib")
    >>> import polars as pl

    We'll imagine an elementary school is taking a pet census.
    The private census data will have two columns: 

    >>> lf_domain = dp.lazyframe_domain([
    ...     dp.series_domain("grade", dp.atom_domain(T=dp.i32)),
    ...     dp.series_domain("pet_count", dp.atom_domain(T=dp.i32))])

    We also need to specify the column we'll be grouping by.

    >>> lf_domain_with_margin = dp.with_margin(
    ...     lf_domain,
    ...     dp.polars.Margin(
    ...         by=[pl.col("grade")],
    ...         invariant="keys",
    ...         max_length=50))

    With that in place, we can plan the Polars computation, using the ``dp`` plugin. 

    >>> plan = (
    ...     pl.LazyFrame(schema={'grade': pl.Int32, 'pet_count': pl.Int32})
    ...     .group_by("grade")
    ...     .agg(pl.col("pet_count").dp.sum((0, 10), scale=1.0)))

    We now have all the pieces to make our measurement function using `make_private_lazyframe`:

    >>> dp_sum_pets_by_grade = dp.m.make_private_lazyframe(
    ...     input_domain=lf_domain_with_margin,
    ...     input_metric=dp.symmetric_distance(),
    ...     output_measure=dp.max_divergence(),
    ...     lazyframe=plan,
    ...     global_scale=1.0)

    It's only at this point that we need to introduce the private data.

    >>> df = pl.from_records(
    ...     [
    ...         [0, 0], # No kindergarteners with pets.
    ...         [0, 0],
    ...         [0, 0],
    ...         [1, 1], # Each first grader has 1 pet.
    ...         [1, 1],
    ...         [1, 1],
    ...         [2, 1], # One second grader has chickens!
    ...         [2, 1],
    ...         [2, 9]
    ...     ],
    ...     schema=['grade', 'pet_count'], orient="row")
    >>> lf = pl.LazyFrame(df)
    >>> results = dp_sum_pets_by_grade(lf).collect()
    >>> print(results.sort("grade")) # doctest: +ELLIPSIS
    shape: (3, 2)
    ┌───────┬───────────┐
    │ grade ┆ pet_count │
    │ ---   ┆ ---       │
    │ i64   ┆ i64       │
    ╞═══════╪═══════════╡
    │ 0     ┆ ...       │
    │ 1     ┆ ...       │
    │ 2     ┆ ...       │
    └───────┴───────────┘

    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_private_lazyframe(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        lazyframe=lazyframe,
        global_scale=global_scale,
        threshold=threshold))
    output.__opendp_dict__ = {
            '__function__': 'then_private_lazyframe',
            '__module__': 'measurements',
            '__kwargs__': {
                'output_measure': output_measure, 'lazyframe': lazyframe, 'global_scale': global_scale, 'threshold': threshold
            },
        }
    return output



def make_private_quantile(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    candidates,
    alpha: float,
    scale: float
) -> Measurement:
    r"""Makes a Measurement the computes the quantile of a dataset.


    Required features: `contrib`

    [make_private_quantile in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_private_quantile.html)

    **Supporting Elements:**

    * Input Domain:   `VectorDomain<AtomDomain<T>>`
    * Output Type:    `MI`
    * Input Metric:   `MO`
    * Output Measure: `T`

    .. end-markdown

    :param input_domain: Uses a tighter sensitivity when the size of vectors in the input domain is known.
    :type input_domain: Domain
    :param input_metric: Either SymmetricDistance or InsertDeleteDistance.
    :type input_metric: Metric
    :param output_measure: Either MaxDivergence or ZeroConcentratedDivergence.
    :type output_measure: Measure
    :param candidates: Potential quantiles to score
    :param alpha: a value in $[0, 1]$. Choose 0.5 for median
    :type alpha: float
    :param scale: the scale of the noise added
    :type scale: float
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # Standardize type arguments.
    T = get_atom(get_type(input_domain)) # type: ignore

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name=None)
    c_candidates = py_to_c(candidates, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=[T]))
    c_alpha = py_to_c(alpha, c_type=ctypes.c_double, type_name="f64")
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")

    # Call library function.
    lib_function = lib.opendp_measurements__make_private_quantile
    lib_function.argtypes = [Domain, Metric, Measure, AnyObjectPtr, ctypes.c_double, ctypes.c_double]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_candidates, c_alpha, c_scale), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_private_quantile',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'candidates': candidates, 'alpha': alpha, 'scale': scale
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_private_quantile(
    output_measure: Measure,
    candidates,
    alpha: float,
    scale: float
):  
    r"""partial constructor of make_private_quantile

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_private_quantile`

    :param output_measure: Either MaxDivergence or ZeroConcentratedDivergence.
    :type output_measure: Measure
    :param candidates: Potential quantiles to score
    :param alpha: a value in $[0, 1]$. Choose 0.5 for median
    :type alpha: float
    :param scale: the scale of the noise added
    :type scale: float
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_private_quantile(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        candidates=candidates,
        alpha=alpha,
        scale=scale))
    output.__opendp_dict__ = {
            '__function__': 'then_private_quantile',
            '__module__': 'measurements',
            '__kwargs__': {
                'output_measure': output_measure, 'candidates': candidates, 'alpha': alpha, 'scale': scale
            },
        }
    return output



def make_randomized_response(
    categories,
    prob: float,
    T: Optional[RuntimeTypeDescriptor] = None
) -> Measurement:
    r"""Make a Measurement that implements randomized response on a categorical value.


    Required features: `contrib`

    [make_randomized_response in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_randomized_response.html)

    **Supporting Elements:**

    * Input Domain:   `AtomDomain<T>`
    * Output Type:    `DiscreteDistance`
    * Input Metric:   `MaxDivergence`
    * Output Measure: `T`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/randomized_response/make_randomized_response.pdf)

    .. end-markdown

    :param categories: Set of valid outcomes
    :param prob: Probability of returning the correct answer. Must be in ``[1/num_categories, 1]``
    :type prob: float
    :param T: Data type of a category.
    :type T: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.enable_features("contrib")
    >>> random_string = dp.m.make_randomized_response(['a', 'b', 'c'], 0.99)
    >>> print('a?', random_string('a'))
    a? ...

    """
    assert_features("contrib")

    # Standardize type arguments.
    T = RuntimeType.parse_or_infer(type_name=T, public_example=get_first(categories))

    # Convert arguments to c types.
    c_categories = py_to_c(categories, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=[T]))
    c_prob = py_to_c(prob, c_type=ctypes.c_double, type_name="f64")
    c_T = py_to_c(T, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_measurements__make_randomized_response
    lib_function.argtypes = [AnyObjectPtr, ctypes.c_double, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_categories, c_prob, c_T), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_randomized_response',
            '__module__': 'measurements',
            '__kwargs__': {
                'categories': categories, 'prob': prob, 'T': T
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def make_randomized_response_bitvec(
    input_domain: Domain,
    input_metric: Metric,
    f: float,
    constant_time: bool = False
) -> Measurement:
    r"""Make a Measurement that implements randomized response on a bit vector.

    This primitive can be useful for implementing RAPPOR.


    Required features: `contrib`

    [make_randomized_response_bitvec in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_randomized_response_bitvec.html)

    **Citations:**

    * [RAPPOR: Randomized Aggregatable Privacy-Preserving Ordinal Response](https://arxiv.org/abs/1407.6981)

    **Supporting Elements:**

    * Input Domain:   `BitVectorDomain`
    * Output Type:    `DiscreteDistance`
    * Input Metric:   `MaxDivergence`
    * Output Measure: `BitVector`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/randomized_response_bitvec/make_randomized_response_bitvec.pdf)

    .. end-markdown

    :param input_domain: BitVectorDomain with max_weight
    :type input_domain: Domain
    :param input_metric: DiscreteDistance
    :type input_metric: Metric
    :param f: Per-bit flipping probability. Must be in $(0, 1]$.
    :type f: float
    :param constant_time: Whether to run the Bernoulli samplers in constant time, this is likely to be extremely slow.
    :type constant_time: bool
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    .. code:: pycon

        >>> import numpy as np
        >>> import opendp.prelude as dp

        >>> dp.enable_features("contrib")

        >>> # Create the randomized response mechanism
        >>> m_rr = dp.m.make_randomized_response_bitvec(
        ...     dp.bitvector_domain(max_weight=4), dp.discrete_distance(), f=0.95
        ... )

        >>> # compute privacy loss
        >>> m_rr.map(1)
        0.8006676684558611

        >>> # formula is 2 * m * ln((2 - f) / f)
        >>> # where m = 4 (the weight) and f = .95 (the flipping probability)

        >>> # prepare a dataset to release, by encoding a bit vector as a numpy byte array
        >>> data = np.packbits(
        ...     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0]
        ... )
        >>> assert np.array_equal(data, np.array([0, 8, 12], dtype=np.uint8))

        >>> # roundtrip: numpy -> bytes -> mech -> bytes -> numpy
        >>> release = np.frombuffer(m_rr(data.tobytes()), dtype=np.uint8)

        >>> # compare the two bit vectors:
        >>> [int(bit) for bit in np.unpackbits(data)]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0]
        >>> [int(bit) for bit in np.unpackbits(release)]
        [...]

    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_f = py_to_c(f, c_type=ctypes.c_double, type_name="f64")
    c_constant_time = py_to_c(constant_time, c_type=ctypes.c_bool, type_name="bool")

    # Call library function.
    lib_function = lib.opendp_measurements__make_randomized_response_bitvec
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, ctypes.c_bool]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_f, c_constant_time), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_randomized_response_bitvec',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'f': f, 'constant_time': constant_time
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_randomized_response_bitvec(
    f: float,
    constant_time: bool = False
):  
    r"""partial constructor of make_randomized_response_bitvec

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_randomized_response_bitvec`

    :param f: Per-bit flipping probability. Must be in $(0, 1]$.
    :type f: float
    :param constant_time: Whether to run the Bernoulli samplers in constant time, this is likely to be extremely slow.
    :type constant_time: bool

    :example:

    .. code:: pycon

        >>> import numpy as np
        >>> import opendp.prelude as dp

        >>> dp.enable_features("contrib")

        >>> # Create the randomized response mechanism
        >>> m_rr = dp.m.make_randomized_response_bitvec(
        ...     dp.bitvector_domain(max_weight=4), dp.discrete_distance(), f=0.95
        ... )

        >>> # compute privacy loss
        >>> m_rr.map(1)
        0.8006676684558611

        >>> # formula is 2 * m * ln((2 - f) / f)
        >>> # where m = 4 (the weight) and f = .95 (the flipping probability)

        >>> # prepare a dataset to release, by encoding a bit vector as a numpy byte array
        >>> data = np.packbits(
        ...     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0]
        ... )
        >>> assert np.array_equal(data, np.array([0, 8, 12], dtype=np.uint8))

        >>> # roundtrip: numpy -> bytes -> mech -> bytes -> numpy
        >>> release = np.frombuffer(m_rr(data.tobytes()), dtype=np.uint8)

        >>> # compare the two bit vectors:
        >>> [int(bit) for bit in np.unpackbits(data)]
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0]
        >>> [int(bit) for bit in np.unpackbits(release)]
        [...]

    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_randomized_response_bitvec(
        input_domain=input_domain,
        input_metric=input_metric,
        f=f,
        constant_time=constant_time))
    output.__opendp_dict__ = {
            '__function__': 'then_randomized_response_bitvec',
            '__module__': 'measurements',
            '__kwargs__': {
                'f': f, 'constant_time': constant_time
            },
        }
    return output



def make_randomized_response_bool(
    prob: float,
    constant_time: bool = False
) -> Measurement:
    r"""Make a Measurement that implements randomized response on a boolean value.


    Required features: `contrib`

    [make_randomized_response_bool in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_randomized_response_bool.html)

    **Supporting Elements:**

    * Input Domain:   `AtomDomain<bool>`
    * Output Type:    `DiscreteDistance`
    * Input Metric:   `MaxDivergence`
    * Output Measure: `bool`

    **Proof Definition:**

    [(Proof Document)](https://docs.opendp.org/en/v0.14.0/proofs/rust/src/measurements/randomized_response/make_randomized_response_bool.pdf)

    .. end-markdown

    :param prob: Probability of returning the correct answer. Must be in ``[0.5, 1]``
    :type prob: float
    :param constant_time: Set to true to enable constant time. Slower.
    :type constant_time: bool
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.enable_features("contrib")
    >>> random_bool = dp.m.make_randomized_response_bool(0.99)
    >>> print('True?', random_bool(True))
    True? ...

    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_prob = py_to_c(prob, c_type=ctypes.c_double, type_name="f64")
    c_constant_time = py_to_c(constant_time, c_type=ctypes.c_bool, type_name="bool")

    # Call library function.
    lib_function = lib.opendp_measurements__make_randomized_response_bool
    lib_function.argtypes = [ctypes.c_double, ctypes.c_bool]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_prob, c_constant_time), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_randomized_response_bool',
            '__module__': 'measurements',
            '__kwargs__': {
                'prob': prob, 'constant_time': constant_time
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


@deprecated(version="0.14.0", reason="use `make_noisy_max` instead")
def make_report_noisy_max_gumbel(
    input_domain: Domain,
    input_metric: Metric,
    scale: float,
    optimize: str = "max"
) -> Measurement:
    r"""Make a Measurement that takes a vector of scores and privately selects the index of the highest score.


    Required features: `contrib`

    [make_report_noisy_max_gumbel in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/measurements/fn.make_report_noisy_max_gumbel.html)

    **Supporting Elements:**

    * Input Domain:   `VectorDomain<AtomDomain<TIA>>`
    * Output Type:    `LInfDistance<TIA>`
    * Input Metric:   `MaxDivergence`
    * Output Measure: `usize`

    .. end-markdown

    :param input_domain: Domain of the input vector. Must be a non-nullable ``VectorDomain``
    :type input_domain: Domain
    :param input_metric: Metric on the input domain. Must be ``LInfDistance``
    :type input_metric: Metric
    :param scale: Scale for the noise distribution
    :type scale: float
    :param optimize: Set to "min" to report noisy min
    :type optimize: str
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_scale = py_to_c(scale, c_type=ctypes.c_double, type_name="f64")
    c_optimize = py_to_c(optimize, c_type=ctypes.c_char_p, type_name="String")

    # Call library function.
    lib_function = lib.opendp_measurements__make_report_noisy_max_gumbel
    lib_function.argtypes = [Domain, Metric, ctypes.c_double, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_scale, c_optimize), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_report_noisy_max_gumbel',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'scale': scale, 'optimize': optimize
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_report_noisy_max_gumbel(
    scale: float,
    optimize: str = "max"
):  
    r"""partial constructor of make_report_noisy_max_gumbel

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_report_noisy_max_gumbel`

    :param scale: Scale for the noise distribution
    :type scale: float
    :param optimize: Set to "min" to report noisy min
    :type optimize: str
    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_report_noisy_max_gumbel(
        input_domain=input_domain,
        input_metric=input_metric,
        scale=scale,
        optimize=optimize))
    output.__opendp_dict__ = {
            '__function__': 'then_report_noisy_max_gumbel',
            '__module__': 'measurements',
            '__kwargs__': {
                'scale': scale, 'optimize': optimize
            },
        }
    return output



def make_user_measurement(
    input_domain: Domain,
    input_metric: Metric,
    output_measure: Measure,
    function,
    privacy_map,
    TO: RuntimeTypeDescriptor = "ExtrinsicObject"
) -> Measurement:
    r"""Construct a Measurement from user-defined callbacks.


    Required features: `contrib`, `honest-but-curious`

    **Why honest-but-curious?:**

    This constructor only returns a valid measurement if for every pair of elements $x, x'$ in `input_domain`,
    and for every pair `(d_in, d_out)`,
    where `d_in` has the associated type for `input_metric` and `d_out` has the associated type for `output_measure`,
    if $x, x'$ are `d_in`-close under `input_metric`, `privacy_map(d_in)` does not raise an exception,
    and `privacy_map(d_in) <= d_out`,
    then `function(x), function(x')` are d_out-close under `output_measure`.

    In addition, `function` must not have side-effects, and `privacy_map` must be a pure function.

    **Supporting Elements:**

    * Input Domain:   `AnyDomain`
    * Output Type:    `AnyMetric`
    * Input Metric:   `AnyMeasure`
    * Output Measure: `AnyObject`

    .. end-markdown

    :param input_domain: A domain describing the set of valid inputs for the function.
    :type input_domain: Domain
    :param input_metric: The metric from which distances between adjacent inputs are measured.
    :type input_metric: Metric
    :param output_measure: The measure from which distances between adjacent output distributions are measured.
    :type output_measure: Measure
    :param function: A function mapping data from ``input_domain`` to a release of type ``TO``.
    :param privacy_map: A function mapping distances from ``input_metric`` to ``output_measure``.
    :param TO: The data type of outputs from the function.
    :type TO: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.enable_features("contrib")
    >>> def const_function(_arg):
    ...     return 42
    >>> def privacy_map(_d_in):
    ...     return 0.
    >>> space = dp.atom_domain(T=int), dp.absolute_distance(int)
    >>> user_measurement = dp.m.make_user_measurement(
    ...     *space,
    ...     output_measure=dp.max_divergence(),
    ...     function=const_function,
    ...     privacy_map=privacy_map
    ... )
    >>> print('42?', user_measurement(0))
    42? 42



    """
    assert_features("contrib", "honest-but-curious")

    # Standardize type arguments.
    TO = RuntimeType.parse(type_name=TO)

    # Convert arguments to c types.
    c_input_domain = py_to_c(input_domain, c_type=Domain, type_name=None)
    c_input_metric = py_to_c(input_metric, c_type=Metric, type_name=None)
    c_output_measure = py_to_c(output_measure, c_type=Measure, type_name="AnyMeasure")
    c_function = py_to_c(function, c_type=CallbackFnPtr, type_name=pass_through(TO))
    c_privacy_map = py_to_c(privacy_map, c_type=CallbackFnPtr, type_name=measure_distance_type(output_measure))
    c_TO = py_to_c(TO, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_measurements__make_user_measurement
    lib_function.argtypes = [Domain, Metric, Measure, CallbackFnPtr, CallbackFnPtr, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_input_domain, c_input_metric, c_output_measure, c_function, c_privacy_map, c_TO), Measurement))
    try:
        output.__opendp_dict__ = {
            '__function__': 'make_user_measurement',
            '__module__': 'measurements',
            '__kwargs__': {
                'input_domain': input_domain, 'input_metric': input_metric, 'output_measure': output_measure, 'function': function, 'privacy_map': privacy_map, 'TO': TO
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output

def then_user_measurement(
    output_measure: Measure,
    function,
    privacy_map,
    TO: RuntimeTypeDescriptor = "ExtrinsicObject"
):  
    r"""partial constructor of make_user_measurement

    .. end-markdown

    .. seealso:: 
      Delays application of ``input_domain`` and ``input_metric`` in :py:func:`opendp.measurements.make_user_measurement`

    :param output_measure: The measure from which distances between adjacent output distributions are measured.
    :type output_measure: Measure
    :param function: A function mapping data from ``input_domain`` to a release of type ``TO``.
    :param privacy_map: A function mapping distances from ``input_metric`` to ``output_measure``.
    :param TO: The data type of outputs from the function.
    :type TO: :py:ref:`RuntimeTypeDescriptor`

    :example:

    >>> dp.enable_features("contrib")
    >>> def const_function(_arg):
    ...     return 42
    >>> def privacy_map(_d_in):
    ...     return 0.
    >>> space = dp.atom_domain(T=int), dp.absolute_distance(int)
    >>> user_measurement = dp.m.make_user_measurement(
    ...     *space,
    ...     output_measure=dp.max_divergence(),
    ...     function=const_function,
    ...     privacy_map=privacy_map
    ... )
    >>> print('42?', user_measurement(0))
    42? 42



    """
    output = _PartialConstructor(lambda input_domain, input_metric: make_user_measurement(
        input_domain=input_domain,
        input_metric=input_metric,
        output_measure=output_measure,
        function=function,
        privacy_map=privacy_map,
        TO=TO))
    output.__opendp_dict__ = {
            '__function__': 'then_user_measurement',
            '__module__': 'measurements',
            '__kwargs__': {
                'output_measure': output_measure, 'function': function, 'privacy_map': privacy_map, 'TO': TO
            },
        }
    return output




"""Release multiple algorithms sequentially."""

from dataclasses import dataclass
from typing import Any, Optional, TYPE_CHECKING

from opendp._internal import _new_pure_function
from opendp._lib import import_optional_dependency
from opendp.combinators import make_adaptive_composition
from opendp.extras.mbi._utilities import (
    Algorithm,
    get_associated_metric,
    get_cardinalities,
)
from opendp.metrics import frame_distance, symmetric_distance
from opendp.mod import (
    FrameDistance,
    LazyFrameDomain,
    Measure,
    Queryable,
    binary_search_param,
)

if TYPE_CHECKING:  # pragma: no cover
    from opendp.extras.polars import Bound


@dataclass(kw_only=True, frozen=True)
class Sequential(Algorithm):
    """Release multiple algorithms sequentially.

    Use this in conjunction with Fixed to ensure specific marginals are estimated.
    """

    algorithms: list[Algorithm]
    """Sequence of algorithms."""
    weights: Optional[list[float]] = None
    """Budget allocation amongst algorithms.
    
    Defaults to equal budget for each algorithm."""

    def __post_init__(self):
        super().__post_init__()

        if not self.algorithms:
            raise ValueError("algorithms must contain at least one element")

        if any(not isinstance(a, Algorithm) for a in self.algorithms):
            raise ValueError(f"algorithms ({self.algorithms}) must be instances of Algorithm")

        if self.weights is not None:
            if len(self.weights) != len(self.algorithms):
                msg = "algorithms and weights must contain the same number of elements"
                raise ValueError(msg)

            if any(w <= 0 for w in self.weights):
                raise ValueError(f"weights ({self.weights}) must be positive")

    def make_marginals(
        self,
        input_domain: LazyFrameDomain,
        input_metric: FrameDistance,
        output_measure: Measure,
        d_in: list["Bound"],
        d_out: float,
        *,
        marginals: dict[tuple[str, ...], Any],
        model: Any,  # MarkovRandomField
    ):
        """Implements a "Sequential" algorithm over ordinal data.

        The algorithm applies each algorithm sequentially.

        :param input_domain: domain of input data
        :param input_metric: how to compute distance between datasets
        :param output_measure: how to measure privacy of release
        :param d_in: distance between adjacent input datasets
        :param d_out: upper bound on the privacy loss
        :param marginals: prior marginal releases
        :param model: warm-start fit of MarkovRandomField
        """
        import_optional_dependency("mbi")
        from mbi import MarkovRandomField  # type: ignore[import-untyped,import-not-found]
        import numpy as np  # type: ignore[import-not-found]

        get_cardinalities(input_domain)
        if input_metric != frame_distance(symmetric_distance()):
            message = f"input_metric ({input_metric}) must be frame_distance(symmetric_distance())"
            raise ValueError(message)
        get_associated_metric(output_measure)

        if not isinstance(model, MarkovRandomField):
            raise ValueError("model must be a MarkovRandomField")

        weights = np.array(self.weights or [1] * len(self.algorithms))

        scale = binary_search_param(
            lambda scale: make_adaptive_composition(
                input_domain, input_metric, output_measure, d_in, d_mids=weights * scale
            ),
            d_in=d_in,
            d_out=d_out,
            T=float,
        )
        d_mids = weights * scale

        def function(
            qbl: Queryable,
        ) -> tuple[dict[tuple[str, ...], Any], MarkovRandomField]:
            all_marginals = marginals.copy()
            current_model = model
            for algo, d_mid in zip(self.algorithms, d_mids):
                m_algo = algo.make_marginals(
                    input_domain,
                    input_metric,
                    output_measure,
                    d_in,
                    d_mid,
                    marginals=all_marginals,
                    model=current_model,
                )

                all_marginals, current_model = qbl(m_algo)

            return all_marginals, current_model

        return make_adaptive_composition(
            input_domain, input_metric, output_measure, d_in, d_mids=d_mids
        ) >> _new_pure_function(function)



import pytest
import opendp.prelude as dp


def test_gaussian_curve():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    meas = dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, 4.))
    profile = meas.map(d_in=1.)
    assert profile.epsilon(delta=0.) == float('inf')
    # see cdp_delta for formula of 0.688 and 0.151

    # cdp_epsilon(rho=(1/4)^2 / 2, delta=1e-3)
    assert profile.epsilon(delta=1e-3) == 0.6880024554878085
    assert profile.epsilon(delta=1.) == 0.
    # cdp_delta(rho=(1/4)^2 / 2, epsilon=0.0)
    assert profile.delta(epsilon=0.) == 0.1508457845622862
    # reuse the constant above
    assert profile.delta(epsilon=0.6880024554878085) == 1e-3

    profile = dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, 4.)).map(d_in=0.0)
    assert profile.epsilon(0.0) == 0.0
    with pytest.raises(dp.OpenDPException):
        profile.epsilon(delta=-0.0)
    with pytest.raises(dp.OpenDPException):
        profile.delta(epsilon=-0.0)
        
    profile = dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, 0.)).map(d_in=1.0)
    assert profile.epsilon(delta=0.0) == float('inf')
    assert profile.epsilon(delta=0.1) == float('inf')
    assert profile.delta(epsilon=0.0) == 1.0
    assert profile.delta(epsilon=0.1) == 1.0

    profile = dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, 0.)).map(d_in=0.0)
    assert profile.epsilon(delta=0.0) == 0.0
    assert profile.epsilon(delta=0.1) == 0.0
    assert profile.delta(epsilon=0.0) == 0.0
    assert profile.delta(epsilon=0.1) == 0.0



def test_gaussian_search():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)

    def make_approx_gauss(scale, delta):
        return dp.c.make_fix_delta(dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, scale)), delta)

    fixed_meas = make_approx_gauss(1., 1e-5)
    ideal_dist = fixed_meas.map(1.)
    print("ideal dist", ideal_dist)
    print("check with ideal dist:", fixed_meas.check(1., ideal_dist))

    print("Standard", dp.binary_search_param(
        lambda s: make_approx_gauss(s, 1e-5),
        d_in=1., d_out=(1., 1e-5)))

def new_make_noise(measure):
    def make_noise(domain, metric, scale):
        return dp.m.make_noise(domain, metric, measure, scale)
    return make_noise

@pytest.mark.parametrize("constructor", [
    dp.m.make_laplace,
    new_make_noise(dp.max_divergence())
])
def test_laplace(constructor):
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    meas = constructor(*input_space, 1)
    assert -50 < meas(0.) < 50
    assert meas.map(1.0) == 1.0


@pytest.mark.parametrize("constructor", [
    dp.m.make_laplace,
    new_make_noise(dp.max_divergence())
])
def test_vector_laplace(constructor):
    input_space = dp.vector_domain(dp.atom_domain(T=float, nan=False)), dp.l1_distance(T=float)
    meas = constructor(*input_space, scale=1.)
    release = meas([0., 0., 0.])
    assert -50 < min(release)
    assert max(release) < 50
    assert meas.map(1.0) == 1.0


def test_gaussian_smoothed_max_divergence():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    meas = dp.c.make_zCDP_to_approxDP(dp.m.make_gaussian(*input_space, scale=10.5))
    print("base gaussian:", meas(100.))

    epsilon = meas.map(d_in=1.).epsilon(delta=.000001)
    print("epsilon:", epsilon)
    assert epsilon > .4


def test_gaussian_zcdp():
    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    meas = input_space >> dp.m.then_gaussian(scale=1.5, MO=dp.ZeroConcentratedDivergence)
    print("base gaussian:", meas(100.))

    rho = meas.map(d_in=1.)
    print("rho:", rho)



def test_vector_gaussian():
    delta = .000001
    input_space = dp.vector_domain(dp.atom_domain(T=float, nan=False)), dp.l2_distance(T=float)
    meas = dp.c.make_fix_delta(
        dp.c.make_zCDP_to_approxDP(
            dp.m.make_gaussian(*input_space, scale=10.5)), delta)
    print("base gaussian:", meas([80., 90., 100.]))
    assert meas.check(1., (0.6, delta))


def test_geometric():
    input_space = dp.atom_domain(T=int), dp.absolute_distance(T=int)
    meas = input_space >> dp.m.then_geometric(scale=2., bounds=(1, 10))
    print("base_geometric in constant time:", meas(100))
    assert meas.check(1, 0.5)
    assert not meas.check(1, 0.49999)

    meas = input_space >> dp.m.then_geometric(scale=2.)
    print("base_geometric:", meas(100))
    assert meas.check(1, 0.5)
    assert not meas.check(1, 0.49999)


def test_discrete_laplace():
    meas = dp.m.make_laplace(dp.atom_domain(T=int), dp.absolute_distance(T=int), scale=2.)
    print("base_discrete_laplace:", meas(100))
    assert meas.check(1, 0.5)
    assert not meas.check(1, 0.49999)

def test_discrete_laplace_cks20():
    meas = dp.m.make_laplace(dp.atom_domain(T=int), dp.absolute_distance(T=int), scale=2.)
    print("base_discrete_laplace:", meas(100))
    assert meas.check(1, 0.5)
    assert not meas.check(1, 0.49999)

def test_vector_discrete_laplace_cks20():
    input_space = dp.vector_domain(dp.atom_domain(T=int)), dp.l1_distance(T=int)
    meas = input_space >> dp.m.then_laplace(scale=2.)
    print("vector base_dl:", meas([100, 10, 12]))
    assert meas.check(1, 0.5)
    assert not meas.check(1, 0.49999)


def test_discrete_laplace_linear():
    input_space = dp.atom_domain(T=int), dp.absolute_distance(T=int)
    meas = dp.m.make_geometric(*input_space, scale=2., bounds=(1, 10))
    print("base_discrete_laplace:", meas(100))
    assert meas.check(1, 0.5)
    assert not meas.check(1, 0.49999)


def test_vector_discrete_laplace():
    input_space = dp.vector_domain(dp.atom_domain(T=int)), dp.l1_distance(T=int)
    meas = dp.m.make_laplace(*input_space, scale=2.)
    print("vector base_dl:", meas([100, 10, 12]))
    assert meas.check(1, 0.5)
    assert not meas.check(1, 0.49999)

def test_discrete_gaussian():
    input_space = dp.atom_domain(T=int), dp.absolute_distance(T=int)
    meas = dp.m.make_gaussian(*input_space, scale=2.)
    print("base_discrete_gaussian:", meas(100))
    assert meas.check(1, 0.5)
    assert meas.check(1, 0.125)

def test_vector_discrete_gaussian():
    input_space = dp.vector_domain(dp.atom_domain(T=int)), dp.l2_distance(T=float)
    meas = dp.m.make_gaussian(*input_space, scale=2.)
    print("vector base_dl:", meas([100, 10, 12]))
    assert meas.check(1., 0.125)
    assert not meas.check(1., 0.124)

def test_make_count_by_ptr():
    meas = (
        dp.space_of(list[str]) >>
        dp.t.then_count_by() >> 
        dp.m.then_laplace_threshold(scale=2., threshold=28)
    )
    print("stability histogram:", meas(["CAT_A"] * 20 + ["CAT_B"] * 10))
    assert meas.map(1) == (0.5, 5.175928103895444e-07)
    assert meas.check(1, (1.0, 1e-6))

def test_randomized_response():
    meas = dp.m.make_randomized_response(categories=["A", "B", "C", "D"], prob=0.75)
    print("randomized response:", meas("A"))
    import math
    assert meas.check(1, math.log(9.))
    assert not meas.check(1, math.log(8.999))


def test_randomized_response_bool():
    meas = dp.m.make_randomized_response_bool(prob=0.75)
    print("randomized response:", meas(True))
    import math
    assert meas.check(1, math.log(3.))
    assert not meas.check(1, math.log(2.999))


def test_gaussian():
    input_space = dp.atom_domain(T=int), dp.absolute_distance(T=float)
    (input_space >> dp.m.then_gaussian(1.))(1)

    input_space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    (input_space >> dp.m.then_gaussian(1.))(1.)

    input_space = dp.vector_domain(dp.atom_domain(T=int)), dp.l2_distance(T=float)
    (input_space >> dp.m.then_gaussian(1.))([1, 2, 3])

    input_space = dp.vector_domain(dp.atom_domain(T=float, nan=False)), dp.l2_distance(T=float)
    (input_space >> dp.m.then_gaussian(1.))([1., 2., 3.])


@pytest.mark.parametrize("measure,d_out", [
    # d_in * 2 / scale = 2
    (dp.max_divergence(), 2), 
    # (d_in * 2 / scale)^2 / 8
    (dp.zero_concentrated_divergence(), 1 / 2)
])
def test_noisy_max(measure, d_out):
    input_domain = dp.vector_domain(dp.atom_domain(T=dp.usize))
    input_metric = dp.linf_distance(T=dp.usize)

    meas = (input_domain, input_metric) >> dp.m.then_noisy_max(measure, 1.)
    # fails with very small probability
    assert meas([0, 0, 20, 40]) == 3  # because score 3 is by far the greatest
    assert meas.map(1) == d_out


@pytest.mark.parametrize("measure,d_out", [
    # (d_in * 2) / scale * 2 = 4
    (dp.max_divergence(), 4), 
    # ((d_in * 2) / scale)^2 / 8 * 2 = 1
    (dp.zero_concentrated_divergence(), 1)
])
def test_noisy_top_k(measure, d_out):
    input_domain = dp.vector_domain(dp.atom_domain(T=dp.usize))
    input_metric = dp.linf_distance(T=dp.usize)

    meas = (input_domain, input_metric) >> dp.m.then_noisy_top_k(measure, 2, 1.)
    # fails with very small probability
    assert meas([0, 0, 20, 40]) == [3, 2]  # because score 3 and then 2 are by far the greatest

    assert meas.map(1) == d_out


def test_alp_histogram():
    counter = dp.t.make_count_by(
        dp.vector_domain(dp.atom_domain(T=str)),
        dp.symmetric_distance())

    alp_meas = counter >> dp.m.then_alp_queryable(
        scale=1.,
        total_limit=24,
        value_limit=24,
    )

    alp_qbl = alp_meas(["A"] * 20 + ["B"] * 10)

    print("Should be high-ish", alp_qbl("A"))
    print("Should be mid-ish", alp_qbl("B"))
    print("Should be low-ish", alp_qbl("C"))
    assert alp_meas.map(1) == 1

def test_randomized_response_bitvec():
    np = pytest.importorskip('numpy')

    m_rr = dp.m.make_randomized_response_bitvec(
        dp.bitvector_domain(max_weight=4), dp.discrete_distance(), f=0.95
    )

    data = np.packbits(
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0]
    )
    assert np.array_equal(data, np.array([0, 8, 12], dtype=np.uint8))

    # roundtrip: bytes -> mech -> numpy
    release = np.frombuffer(m_rr(data), dtype=np.uint8)

    print("np.unpackbits: data vs. release", np.unpackbits(data), np.unpackbits(release))
    # epsilon is 2 * m * ln((2 - f) / f)
    # where m = 4 and f = .95
    assert m_rr.map(1) == 0.8006676684558611

def test_laplace_threshold_int():
    domain = dp.map_domain(dp.atom_domain(T=str), dp.atom_domain(T=int))
    metric = dp.l01inf_distance(dp.absolute_distance(T=int))
    meas = dp.m.make_laplace_threshold(domain, metric, scale=2., threshold=28)
    release = meas({str(i): i * 10 for i in range(10)})
    # 0 + noise is likely not over 28
    assert "0" not in release
    # 90 + noise is likely over 28
    assert "9" in release
    # delta is the mass of the right tail of integer laplace greater than 28 with conservative arithmetic
    # 5 = 10 / 2 = Δ / σ
    assert meas.map((1, 10, 10)) == (5.0, 4.659221997116436e-05)

def test_laplace_threshold_float():
    domain = dp.map_domain(dp.atom_domain(T=str), dp.atom_domain(T=float, nan=False))
    metric = dp.l01inf_distance(dp.absolute_distance(T=float))
    meas = dp.m.make_laplace_threshold(domain, metric, scale=2., threshold=28.)
    release = meas({str(i): i * 10.0 for i in range(10)})
    # 0 + noise is likely not over 28
    assert "0" not in release, release
    # 90 + noise is likely over 28
    assert "9" in release, release
    print(meas.map((1, 10, 10)))
    # delta is the mass of the right tail of continuous laplace greater than 28 with conservative arithmetic
    # the looser continuous laplace is used because discrete laplace tail bound is not computable with large constants
    assert meas.map((1, 10, 10)) == (5.0, 6.17049020433802e-05)

def test_gaussian_threshold_int():
    domain = dp.map_domain(dp.atom_domain(T=str), dp.atom_domain(T=int))
    metric = dp.l02inf_distance(dp.absolute_distance(T=int))
    meas = dp.m.make_gaussian_threshold(domain, metric, scale=2., threshold=28)
    release = meas({str(i): i * 10 for i in range(10)})
    # 0 + noise is likely not over 28
    assert "0" not in release, release
    # 90 + noise is likely over 28
    assert "9" in release, release
    # delta is the mass of the right tail of continuous gaussian greater than 28 with conservative arithmetic
    # the continuous gaussian is used because discrete gaussian tail bound is not analytic
    # 12.5 = (10 / 2)^2 / 2 = (Δ / σ)^2 / 2
    assert meas.map((1, 10, 10)) == (12.5, 1.1102230246251565e-16)

def make_noise_threshold_zCDP(domain, metric, scale, threshold):
    measure = dp.approximate(dp.zero_concentrated_divergence())
    return dp.m.make_noise_threshold(domain, metric, measure, scale, threshold)

@pytest.mark.parametrize("constructor", [
    dp.m.make_gaussian_threshold,
    make_noise_threshold_zCDP
])
def test_gaussian_threshold_float(constructor):
    domain = dp.map_domain(dp.atom_domain(T=str), dp.atom_domain(T=float, nan=False))
    metric = dp.l02inf_distance(dp.absolute_distance(T=float))
    meas = constructor(domain, metric, scale=2., threshold=28.)
    release = meas({str(i): i * 10.0 for i in range(10)})
    # 0 + noise is likely not over 28
    assert "0" not in release, release
    # 90 + noise is likely over 28
    assert "9" in release, release
    # delta is the mass of the right tail of continuous gaussian greater than 28 with conservative arithmetic
    # the continuous gaussian is used because discrete gaussian tail bound is not analytic
    # 12.5 = (10 / 2)^2 / 2 = (Δ / σ)^2 / 2
    assert meas.map((1, 10, 10)) == (12.5, 1.1102230246251565e-16)


def test_canonical_noise():
    space = dp.atom_domain(T=float, nan=False), dp.absolute_distance(T=float)
    m_cnd = space >> dp.m.then_canonical_noise(d_in=1., d_out=(1., 1e-6))

    assert m_cnd.map(1.) == (1.0, 1e-6)
    # just check that it runs
    assert isinstance(m_cnd(0.), float)



# type: ignore
def get_rounding_distance() -> RBig:
    k_min = get_min_k(T) # |\label{line:get-min-k}|
    if k < k_min: # |\label{line:check-k}|
        raise f"k ({k}) must not be smaller than {k_min}"

    # input has granularity 2^{k_min} (subnormal float precision)
    input_gran = x_mul_2k(RBig.ONE, k_min) # |\label{line:input-gran}|

    # discretization rounds to the nearest 2^k
    output_gran = x_mul_2k(RBig.ONE, k) # |\label{line:output-gran}|

    # the worst-case increase in sensitivity due to discretization is
    #     the range, minus the smallest step in the range
    distance = output_gran - input_gran # |\label{line:distance}|

    # rounding may occur on all vector elements
    if not distance.is_zero(): # |\label{line:zero-distance}|
        if size is None: # |\label{line:unknown-size}|
            raise "domain size must be known if discretization is not exact"
        
        match P:
            case 1:
                distance *= RBig.from_(size)
            case 2:
                distance *= RBig.try_from(f64.inf_cast(size).inf_sqrt())
            case _:
                raise f"norm ({P}) must be one or two"
        
    return distance


# type: ignore
class ZExpFamily:
    def make_noise(
        self, input_space: tuple[AtomDomain[IBig], AbsoluteDistance[RBig]]
    ) -> Measurement[AtomDomain[IBig], IBig, AbsoluteDistance[RBig], MO]:
        t_vec = make_vec(input_space)  # |\label{line:make-vec}|
        m_noise = self.make_noise(t_vec.output_space())  # |\label{line:make-noise}|

        return t_vec >> m_noise >> then_index_or_default(0)  # |\label{line:post}|



# type: ignore
def make_expr_datetime_component(
    input_domain: ExprDomain,
    input_metric: M,
    expr: Expr,
) -> Transformation:
    match expr:  # `\label{match-strategy}`
        case Expr.Function(input=inputs, function=FunctionExpr.TemporalExpr(temporal_function)):
            pass
        case _:
            raise ValueError("expected datetime component expression")
    
    to_dtype, _ = match_datetime_component(temporal_function)  # `\label{match-dt}`

    # raises an error if there is not exactly one input
    input, = inputs  # `\label{one-input}`

    t_prior = input.make_stable(input_domain, input_metric)  # `\label{t-prior}`
    middle_domain, middle_metric = t_prior.output_space()

    in_dtype = middle_domain.column.dtype
    if in_dtype not in {DataType.Time, DataType.Datetime, DataType.Date}:  # `\label{check-dtype}`
        raise ValueError("expected a temporal input type")
    
    output_domain = middle_domain.clone()  # `\label{output-domain}`
    output_domain.column.set_dtype(to_dtype) # `\label{domain-dtype}`

    def function(expr: Expr) -> Expr:
        return Expr.Function(
            input=[expr],
            function=FunctionExpr.TemporalExpr(temporal_function),
            options=FunctionOptions(
                collect_groups=ApplyOptions.ElementWise,
            ),
        )

    return t_prior >> Transformation.new(  # `\label{t-component}`
        middle_domain,
        output_domain,
        Function.then_expr(function),
        middle_metric,
        middle_metric,
        StabilityMap.new(lambda d_in: d_in),
    )


# type: ignore
# analogous to impl MakeNoise<VectorDomain<AtomDomain<T>>, L1Distance<T>, MO> for ConstantTimeGeometric<T> in Rust
class ConstantTimeGeometric:
    def make_noise(
        self, input_space: tuple[DI, MI]
    ) -> Measurement[DI, DI_Carrier, MI, MO]:
        input_domain, input_metric = input_space
        scale, (lower, upper) = self.scale, self.bounds
        if lower > upper: # |\label{line:check-bounds}|
            raise "lower may not be greater than upper"

        distribution = ZExpFamily(scale=RBig.from_f64(scale))
        output_measure = MO.default()

        privacy_map = distribution.noise_privacy_map(
            L1Distance.default(), output_measure
        )

        p = (1.0).neg_inf_sub((-scale.recip()).inf_exp()) # |\label{line:prob-check}|
        if not (0.0 < p <= 1.0):
            raise f"Probability of termination p ({p}) must be in (0, 1]. This is likely because the noise scale is so large that conservative arithmetic causes p to go negative"

        def function(arg: Vec[T]) -> Vec[T]:
            return [
                sample_discrete_laplace_linear(v, scale, (lower, upper)) for v in arg
            ]

        return Measurement.new(
            input_domain,
            Function.new_fallible(function),
            input_metric,
            output_measure,
            PrivacyMap.new_fallible(lambda d_in: privacy_map.eval(RBig.try_from(d_in))),
        )



# type: ignore
def new_fully_adaptive_composition_queryable(
    input_domain: DI,
    input_metric: MI,
    output_measure: MO,
    data: DI_Carrier,
) -> OdometerQueryable[Measurement[DI, MI, MO, TO], TO, MO_Distance]:

    require_sequentiality = matches(
        output_measure.composability(Adaptivity.FullyAdaptive),
        Composability.Sequential
    )

    privacy_maps = []  # Vec<PrivacyMap<MI, MO>> `\label{mutable-state}`

    def transition(  # `\label{transition}`
        self_: OdometerQueryable[Measurement[DI, MI, MO, TO], TO, MO_Distance],
        query: Query[OdometerQuery[Measurement[DI, MI, MO, TO]]],
    ):
        # this queryable and wrapped children communicate via an AskPermission query
        # defined here, where no-one else can access the type
        @dataclass
        class AskPermission:
            id: usize

        match query:
            # evaluate external invoke query
            case Query.External(
                OdometerQuery.Invoke(meas)
            ):  # `\label{query-wrapper-def}`
                assert_elements_match(  # `\label{domain-check}`
                    DomainMismatch, input_domain, meas.input_domain
                )

                assert_elements_match(  # `\label{metric-check}`
                    MetricMismatch, input_metric, meas.input_metric
                )

                assert_elements_match(  # `\label{measure-check}`
                    MeasureMismatch, output_measure, meas.output_measure
                )

                enforce_sequentiality = False

                if require_sequentiality:
                    # when the output measure doesn't allow concurrent composition,
                    # wrap any interactive queryables spawned.
                    # This way, when the child gets a query it sends an AskPermission query
                    # to this parent queryable, giving this sequential odometer queryable
                    # a chance to deny the child permission to execute
                    child_id = privacy_maps.len()

                    def callback():
                        nonlocal enforce_sequentiality
                        if enforce_sequentiality:
                            return self_.eval_internal(AskPermission(child_id))
                        else:
                            return ()

                    seq_wrapper = Wrapper.new_recursive_pre_hook(  # `\label{pre-hook}`
                        callback
                    )
                else:
                    seq_wrapper = None

                answer = meas.invoke_wrap(data, seq_wrapper)  # `\label{invoke}`

                enforce_sequentiality = True
                
                # We've now increased our privacy spend.
                # This is our only state modification
                privacy_maps.push(meas.privacy_map)  # `\label{child-privacy-map}`

                return Answer.External(OdometerAnswer.Invoke(answer))

            # evaluate external privacy loss query
            case Query.External(OdometerQuery.PrivacyLoss(d_in)):
                d_mids = [m.eval(d_in) for m in privacy_maps]
                d_out = output_measure.compose(d_mids)
                return Answer.External(OdometerAnswer.Map(d_out))

            case Query.Internal(query):
                # Check if the query is from a child queryable
                #     who is asking for permission to execute
                if isinstance(query, AskPermission):  # `\label{ask-permission-handler}`
                    # deny permission if the sequential odometer has moved on
                    if query.id + 1 != privacy_maps.len():
                        raise ValueError("sequential odometer has received a new query")

                    # otherwise, return Ok to approve the change
                    return Answer.internal(())

                raise ValueError("query not recognized")

    return Queryable.new(transition)  # `\label{queryable}`



# type: ignore
def score_candidates_map(alpha_num, alpha_den, known_size) -> Callable[[int], int]:
    def stability_map(d_in):
        if known_size:
            return T.inf_cast(d_in // 2).inf_mul(alpha_den)
        else:
            abs_dist_const: u64 = max(alpha_num, alpha_den - alpha_num)  # `\label{sub}`
            return T.exact_int_cast(d_in).alerting_mul(abs_dist_const)

    return stability_map



# type: ignore
def match_truncations(
    plan: DslPlan, identifier: Expr
) -> tuple[DslPlan, Vec[Truncation], Vec[Bound]]:
    truncations = []
    bounds = []

    allowed_keys = match_group_by_truncation(plan, identifier) # |\label{line:match_group_by_truncation}|
    if allowed_keys:
        input, truncate, new_bound = allowed_keys
        plan = input # |\label{line:match_group_by_truncation_start}|
        truncations.append(truncate)
        bounds.append(new_bound) # |\label{line:match_group_by_truncation_end}|
        allowed_keys = new_bound.by # |\label{line:allowed_keys}|

    # match until not a filter truncation
    while isinstance(plan, Truncation.Filter): # |\label{line:match_filter_truncation}|
        input, predicate = plan.input, plan.predicate
        new_bounds = match_truncation_predicate(predicate, identifier) # |\label{line:match_truncation_predicate}|
        if new_bounds is None:
            break

        # When filter truncation is behind a groupby truncation,
        # if the groupby group keys don't cover the filter truncation keys,
        # then the groupby aggs can overwrite the filter truncation keys,
        # invalidating the filter truncation bounds.
        if allowed_keys is not None: # |\label{line:allowed_keys_check}|
            for bound in new_bounds:
                if not bound.by.is_subset(allowed_keys):
                    raise f"Filter truncation keys ({bound.by}) must be a subset of groupby truncation keys ({allowed_keys})."

        plan = input # |\label{line:match_filter_truncation_start}|
        truncations.append(Truncation.Filter(predicate))
        bounds.extend(new_bounds) # |\label{line:match_filter_truncation_end}|

    # just for better error messages, no privacy implications
    if match_group_by_truncation(plan, identifier) is not None: # |\label{line:match_group_by_truncation_check}|
        raise Exception("groupby truncation must be the last truncation in the plan.")

    # since the parse descends to the source,
    # truncations and bounds are in reverse order
    truncations.reverse() # |\label{line:truncations_reverse}|
    bounds.reverse() # |\label{line:bounds_reverse}|

    return plan, truncations, bounds



# type: ignore
def match_truncation_predicate(
    predicate: Expr, identifier: Expr
) -> Optional[Vec[Bound]]:
    if (
        isinstance(predicate, FunctionExpr)
        and predicate.function == BooleanFunction.AllHorizontal
    ):
        # propagate errors
        bounds = [
            match_truncation_predicate(expr, identifier) for expr in predicate.input
        ]

        # propagate nones
        if not all(bounds): # |\label{line:check_all}|
            return None
        
        # appears to differ from Rust, but is equivalent 
        # because options don't need to be flattened in Python
        return bounds

    elif isinstance(predicate, BinaryExpr) and predicate.op == Operator.And:
        left = match_truncation_predicate(predicate.left, identifier)
        right = match_truncation_predicate(predicate.right, identifier)
        if left is None or right is None:  # |\label{line:check_both}|
            return None
        return left + right

    elif isinstance(predicate, BinaryExpr):
        left, right = predicate.left, predicate.right
        if predicate.op == Operator.Lt:
            over, threshold, offset = left, right, 0
        elif predicate.op == Operator.LtEq:
            over, threshold, offset = left, right, 1
        elif predicate.op == Operator.Gt:
            over, threshold, offset = right, left, 0
        elif predicate.op == Operator.GtEq:
            over, threshold, offset = right, left, 1
        else:
            return None

        if not isinstance(over, Expr.Window): # |\label{line:check_over}|
            return None

        threshold_value = literal_value_of(threshold, u32)
        if threshold_value is None:
            raise ValueError(
                f"literal value for truncation threshold ({threshold}) must be representable as a u32"
            )

        # account for distinction between gt and ge
        threshold_value = threshold_value.inf_add(offset) # |\label{line:threshold_value}|

        num_groups = match_num_groups_predicate(
            over.function, over.partition_by, identifier, threshold_value
        )
        per_group = match_per_group_predicate(
            over.function, over.partition_by, identifier, threshold_value
        )

        if num_groups is None and per_group is None: # |\label{line:check_bounds}|
            raise ValueError(
                f"expected a predicate that limits per_group contributions (via int_range) or num_groups contributions (via rank). Found {over.function}"
            )

        return [num_groups or per_group] # |\label{line:return}|
